#!/usr/bin/env python3
'''
Created on Nov 23, 2018

@author: ncarrier
'''
import re
import collections
import sys

# required for (at least) re.fullmatch
assert sys.version_info >= (3, 4), "Python version above 3.4 is required"


class VersionNumber(object):
    '''
    Allows to validate and parse product version numbers
    '''
    ALPHA = "alpha"
    BETA = "beta"
    RC = "rc"
    PUBLIC = ""

    __type_suffix = r"(?P<type>alpha|beta|rc)(?P<increment>[1-9][0-9]*)"
    __git_suffix = r"(?P<commits>[1-9][0-9]*)-g(?P<hash>[a-f0-9]{12})"
    # a number which can be 0, but otherwise, without leading zeroes
    __number = r"0|[1-9][0-9]*"
    # this corresponds to the scheme, versions provided to the customers must
    # follow
    __core_version = (r"(?P<major>" + __number + ")\."
                      "(?P<minor>" + __number + ")\."
                      "(?P<patch>" + __number + ")(?P<special>[a-z]?)")
    # this corresponds to the version naming, internal to Orolia
    __version = (r"(?P<core_version>" + __core_version + r")" +
                 r"(-(?P<internal>" + __type_suffix + r"))?" +
                 r"(-(?P<git>" + __git_suffix + r"))?")
    __target = r"[a-z]+"
    __target_version = (r"(?P<target>" + __target + ")-" +
                        r"(?P<version>" + __version + r")")
    # each fields is associated to a conversion function and a default value
    # note that some default values are useless since the corresponding field
    # can't be omitted, for example, major, minor...
    __version_fields = {"type": (str, PUBLIC),
                        "increment": (int, 0),
                        "commits": (int, 0),
                        "hash": (str, ""),
                        "major": (int, 0),
                        "minor": (int, 0),
                        "patch": (int, 0),
                        "special": (str, ""),
                        "core_version": (str, ""),
                        "internal": (str, ""),
                        "git": (str, "")}
    __version_structure = collections.namedtuple("version_structure",
                                                 __version_fields.keys())

    def __init__(self):
        '''
        Constructor
        '''
        self.__regexes = {
            "type_suffix": re.compile(self.__type_suffix),
            "git_suffix": re.compile(self.__git_suffix),
            "core_version": re.compile(self.__core_version),
            "version": re.compile(self.__version),
            "target_version": re.compile(self.__target_version)
        }

    def is_valid(self, regex_name, string):
        return self.__regexes[regex_name].fullmatch(string)

    def is_type_suffix_valid(self, type_suffix):
        return self.is_valid("type_suffix", type_suffix)

    def is_git_suffix_valid(self, git_suffix):
        return self.is_valid("git_suffix", git_suffix)

    def is_core_version_valid(self, core_version):
        return self.is_valid("core_version", core_version)

    def is_version_valid(self, version):
        return self.is_valid("version", version)

    def is_target_version_valid(self, target_version):
        return self.is_valid("target_version", target_version)

    @staticmethod
    def get_version_pattern():
        return VersionNumber.__version

    def get_version(self, version):
        '''
        returns a named tuple representing a parsed version number.
        see __version_fields.keys() for fields
        '''
        match = self.__regexes["version"].match(version)
        vf = self.__version_fields
        vd = {f: vf[f][0](match.group(f) or vf[f][1]) for f in vf}

        return self.__version_structure(**vd)


if __name__ == "__main__":
    print(VersionNumber.get_version_pattern())
